/*********** 
# Copyright (c) 2017, Xilinx, Inc. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software
# without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.
#
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT,INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO,PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY,OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
************/

/******************************************************************************
    SDSoC example with asymmetric multiprocessing with MB
    
    This is simple matrix multiplication example to showcase effectiveness
    of using array partition feature of SDSoC.

*******************************************************************************/

#include <stdio.h>
#include <iostream>
#include <cstdio>
#include <stdlib.h>
#include <xil_io.h>
#include <unistd.h>
#include "xparameters.h"
#include "xmutex.h"

#include "matmul.h"
#include "sds_utils.h"

#include "mem_map.h"

#define MUTEX_DEVICE_ID		XPAR_MUTEX_0_IF_0_DEVICE_ID

/*
 * The Mutex number on which the example is run
 */
#define MUTEX_NUM 		0

/*
 * Mutex instance
 */
XMutex Mutex;

void start_Microblaze() {
    // file handles
    int valuefd, exportfd, directionfd;

    // The GPIO has to be exported to be able to see it
    // in sysfs

    exportfd = open("/sys/class/gpio/export", O_WRONLY);
    if (exportfd < 0)
    {
        printf("Cannot open GPIO to export\n");
        exit(1);
    }

    write(exportfd, "960", 4);
    close(exportfd);

    // Update the direction of the GPIO to be an output

    directionfd = open("/sys/class/gpio/gpio960/direction", O_RDWR);
    if (directionfd < 0)
    {
        printf("Cannot open GPIO direction it\n");
        exit(1);
    }

    write(directionfd, "out", 4);
    close(directionfd);

    // Get the GPIO value ready to be toggled

    valuefd = open("/sys/class/gpio/gpio960/value", O_RDWR);
    if (valuefd < 0)
    {
        printf("Cannot open GPIO value\n");
        exit(1);
    }

    // reset Microblaze
    write(valuefd,"0", 2);

    printf("Reset for 1 sec. ... ");
    sleep(1);
    printf("Done.\n");

    // release MB reset
    write(valuefd,"1", 2);

    close(valuefd);

    //release exported gpio pin
    exportfd = open("/sys/class/gpio/unexport", O_WRONLY);
    if (exportfd < 0)
    {
        printf("Cannot open GPIO to unexport\n");
        exit(1);
    }

    write(exportfd, "960", 4);
    close(exportfd);
    return;
}

// Software Matrix Multiplication 
void matmul(int *C, int *A, int *B, int M) {
    for (int k = 0; k < M; k++) {
        for (int j = 0; j < M; j++) {
            for (int i = 0; i < M; i++) {
                C[k * M + j] += A[k * M + i] * B[i * M + j];
            }
        }
    }
}

// Verify Software and Hardware Results
bool verify(int *gold, int *output, int size) {
    for (int i = 0; i < size; i++) {
        if (output[i] != gold[i]) {
            printf("Mismatch %d: gold: %d hardware: %d\n", i, gold[i], output[i]);
            return false;
        }
    }
    return true;
}

int main(int argc, char **argv) {

	/*
	 * Init mutex and start microblaze
	 */
	XMutex_Config *ConfigPtr;
	XStatus Status;
	u32 TimeoutCount = 0;
	u32 val;

    u32 iiii=0;
    u32 value;

	int memfd;
	memfd = open("/dev/mem", O_RDWR | O_SYNC);
	if (memfd == -1) {
		printf("Can't open /dev/mem.\n");
		exit(0);
	}
	printf("/dev/mem opened.\n");

	/*
	 * Lookup configuration data in the device configuration table.
	 * Use this configuration info down below when initializing this
	 * driver instance.
	 */

	ConfigPtr = XMutex_LookupConfig(MUTEX_DEVICE_ID);
	if (ConfigPtr == (XMutex_Config *) NULL) {
		return XST_FAILURE;
	}

	/*
	 * Map Mutex memory to userspace
	 */
	mem_map_t mutex_map;

	if (map(memfd, (off_t) ConfigPtr->BaseAddress, &mutex_map)) {
		printf("Mutex mapping to userspace failed\r\n");
		exit(1);
	}

	ConfigPtr->BaseAddress = mutex_map.dev_base;

	/*
	 * Perform the rest of the initialization.
	 */
	Status = XMutex_CfgInitialize(&Mutex, ConfigPtr, ConfigPtr->BaseAddress);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}

	XMutex_SetUser(&Mutex, MUTEX_NUM, 0);
	XMutex_Lock(&Mutex, MUTEX_NUM);

	/*
	 * Map mb memory to userspace
	 */
	mem_map_t mb_map;

	if (maps(memfd, 0x28000000, &mb_map, 0x8000000)) {
		printf("Mb_map mapping to userspace failed\r\n");
		exit(1);
	}

	sds_mmap((void *) 0x28000000, 0x8000000, (void *) mb_map.dev_base);

	/*
	 * Write firmware for the EdkDSP at mapped 0x28000000
	 * It is up to 64 firmware programs, each 16kbytes = 4096 uint32 values
	 */

	printf("ARMCPU0: Write firmware ... \n\r");

	FILE *fr0, *fr1, *fr2, *fr3;

	/*
     * open input files to read
     */
    printf("ARMCPU0: Open  input file f0.dec ... ");
    if ( (fr0 = fopen("./f0.dec", "r")) == NULL ) {
        printf("FAILED\n\r");
        return -1;
    }
    printf("OK\n\r");

    printf("ARMCPU0: Open  input file f1.dec ... ");
    if ( (fr1 = fopen("./f1.dec", "r")) == NULL ) {
        printf("FAILED\n\r");
        return -1;
    }
    printf("OK\n\r");

	printf("ARMCPU0: Open  input file f2.dec ... ");
    if ( (fr2 = fopen("./f2.dec", "r")) == NULL ) {
        printf("FAILED\n\r");
        return -1;
    }
    printf("OK\n\r");

    printf("ARMCPU0: Open  input file f3.dec ... ");
    if ( (fr3 = fopen("./f3.dec", "r")) == NULL ) {
        printf("FAILED\n\r");
        return -1;
    }
    printf("OK\n\r");

    iiii = 0;
    value = 139266;
    while (fscanf(fr0,"%d",&value)!=EOF){
    	Xil_Out32(mb_map.dev_base + 4*(0*4096 + iiii), value);
    	iiii++;
    }
    value=139266;
	for (; iiii < 4096; iiii++) {
		Xil_Out32(mb_map.dev_base + 4*(0*4096 + iiii), value);
	}

    iiii = 0;
    value = 139266;
    while (fscanf(fr1,"%d",&value)!=EOF){
    	Xil_Out32(mb_map.dev_base + 4*(1*4096 + iiii), value);
    	iiii++;
    }
    value=139266;
	for (; iiii < 4096; iiii++) {
		Xil_Out32(mb_map.dev_base + 4*(1*4096 + iiii), value);
	}

	iiii = 0;
    value = 139266;
    while (fscanf(fr2,"%d",&value)!=EOF){
    	Xil_Out32(mb_map.dev_base + 4*(2*4096 + iiii), value);
    	iiii++;
    }
    value=139266;
	for (; iiii < 4096; iiii++) {
		Xil_Out32(mb_map.dev_base + 4*(2*4096 + iiii), value);
	}
    
	iiii = 0;
    value = 139266;
    while (fscanf(fr3,"%d",&value)!=EOF){
    	Xil_Out32(mb_map.dev_base + 4*(3*4096 + iiii), value);
    	iiii++;
    }
    value=139266;
	for (; iiii < 4096; iiii++) {
		Xil_Out32(mb_map.dev_base + 4*(3*4096 + iiii), value);
	}

	/*
	 * close input files
	 */
	printf("ARMCPU0: Close input file f0.dec ... ");
	if ( fclose(fr0) == EOF ) {
		printf("FAILED\n\r");
		return -1;
	}
	printf("OK\n\r");

	printf("ARMCPU0: Close input file f1.dec ... ");
	if ( fclose(fr1) == EOF ) {
		printf("FAILED\n\r");
		return -1;
	}
	printf("OK\n\r");

	printf("ARMCPU0: Close input file f2.dec ... ");
	if ( fclose(fr2) == EOF ) {
		printf("FAILED\n\r");
		return -1;
	}
	printf("OK\n\r");

	printf("ARMCPU0: Close input file f3.dec ... ");
	if ( fclose(fr3) == EOF ) {
		printf("FAILED\n\r");
		return -1;
	}
	printf("OK\n\r");

	printf("ARMCPU0: Write firmware Done. \n\r");

	/*
	 * Initial waiting loop for Microblaze mapped at 0x28100000
	 */

#if (1)
	printf("ARMCPU0: place 0x28100000 at start of MB0 vectors\n\r");
	Xil_Out32(mb_map.dev_base + 0x00100000, 0xb8000000);
#endif

	/*
	 * Remove Microblaze reset via system control of an emio pin
	 */

	start_Microblaze();

	printf("ARMCPU0: MB0 reset removed, ARM waiting ... ");

	/*
	 * Wait for response from Microblaze via user value at mutex 0
	 */
	do {
		XMutex_GetUser(&Mutex, MUTEX_NUM, &val);
	} while (val == 0);

	printf("Done.\n\r");

	printf("ARMCPU0: MB0 indicates - running ...\n\r");

    bool test_passed;
    static const int columns = MAX_SIZE;
    static const int rows = MAX_SIZE;

    // Allocate buffers using sds_alloc
    int *A    = (int *) sds_alloc(sizeof(int) * columns * rows);
    int *B    = (int *) sds_alloc(sizeof(int) * columns * rows);
    int *C    = (int *) sds_alloc(sizeof(int) * columns * rows);
    
    // Allocate output buffer for software solution
    int *gold = (int *) malloc(sizeof(int) * columns * rows);

    // Check for failed memory allocation
    if((A == NULL) || (B == NULL) || (C == NULL) || (gold == NULL)){
        std::cout << "TEST FAILED : Failed to allocate memory" << std::endl;
        return -1;
    }

    // Data Initialization
    for(int i = 0; i < columns * rows; ++i) {
        A[i] = i;
        B[i] = i + i;
        gold[i] = 0;
        C[i] = 0;
    }

    sds_utils::perf_counter hw_ctr, sw_ctr;

    sw_ctr.start();
    //Launch the Software Solution
    matmul(gold, A, B, columns);
    sw_ctr.stop();

    hw_ctr.start();
    //Launch the Hardware Solution
    matmul_partition_accel(A, B, C, columns); 
    hw_ctr.stop();

    // Compare the results of hardware to the simulation
    test_passed = verify(gold, C, columns * rows);
    
    uint64_t sw_cycles = sw_ctr.avg_cpu_cycles();
    uint64_t hw_cycles = hw_ctr.avg_cpu_cycles();
    double speedup = (double) sw_cycles / (double) hw_cycles;

    std::cout << "Number of CPU cycles running application in software: "
                << sw_cycles << std::endl;
    std::cout << "Number of CPU cycles running application in hardware: "
                << hw_cycles << std::endl;
    std::cout << "Speed up: " << speedup << std::endl;
    
    sds_free(A);
    sds_free(B);
    sds_free(C);
    free(gold);
    
    std::cout << "TEST " << (test_passed ? "PASSED" : "FAILED") << std::endl; 

	/*
	 *  unmap the memory before exiting
	 */
	if (munmap(mb_map.base, MAP_SIZE) == -1) {
		printf("Can't unmap memory from user space.\n");
		exit(0);
	}

	if (munmap(mutex_map.base, MAP_SIZE) == -1) {
		printf("Can't unmap memory from user space.\n");
		exit(0);
	}

	close(memfd);

    return (test_passed ? 0 : -1);
}
